from unittest import TestCase

from injecty import get_impls, get_instances, get_default_impl, get_new_default_instance
from tests.models import Foo, Bang, Bar, Zap


class TestInjecty(TestCase):
    def test_get_impls(self):
        impls = get_impls(Foo)
        self.assertEqual(impls, [Bang, Bar, Zap])

    def test_get_instances(self):
        instances = [i.__class__ for i in get_instances(Foo)]
        expected = [Bang, Bar, Zap]
        self.assertEqual(instances, expected)

    def test_get_default_impl(self):
        impl = get_default_impl(Foo)
        self.assertEqual(impl, Bang)

    def test_get_new_default_instance(self):
        instance = get_new_default_instance(Foo)
        self.assertIsInstance(instance, Bang)



import importlib
import pkgutil
from dataclasses import dataclass, field
from typing import (
    Dict,
    Type,
    Set,
    List,
    Callable,
    Optional,
    Any,
    TypeVar,
    get_type_hints,
)

_CONFIG_MODULE_PREFIX = "injecty_config"
T = TypeVar("T")


@dataclass
class InjectyContext:
    _impls: Dict[Type[T], Set[Type[T]]] = field(default_factory=dict)

    def register_impl(
        self, base: Type[T], impl: Type[T], check_type: bool = True
    ) -> bool:
        if check_type and not issubclass(impl, base):
            raise TypeError(impl)
        impls = self._impls.get(base)
        if not impls:
            impls = self._impls[base] = set()
        if impl in impls:
            return False
        impls.add(impl)
        return True

    def register_impls(
        self, base: Type[T], impls: List[Type[T]], check_type: bool = True
    ) -> bool:
        result = True
        for impl in impls:
            result &= self.register_impl(base, impl, check_type)
        return result

    def deregister_impl(self, base: Type[T], impl: Type[T]) -> bool:
        impls = self._impls.get(base)
        if impls:
            if impl in impls:
                impls.remove(impl)
                return True
        return False

    def get_impls(
        self,
        base: Type[T],
        sort_key: Optional[Callable[[Type[T]], Any]] = None,
        reverse: bool = False,
        permit_no_impl: bool = False,
    ) -> List[Type]:
        impls = self._impls.get(base)
        if not impls:
            if permit_no_impl:
                return []
            raise ValueError(f"no_implementation_for:{base}")
        result = list(impls)
        if sort_key is None:
            priority = get_type_hints(base).get("priority", None)
            priority = getattr(priority, "__name__", priority)
            if priority == "int":
                sort_key = _priority_sort_key
                reverse = True
        if sort_key:
            # noinspection PyTypeChecker
            result.sort(key=sort_key, reverse=reverse)
        return result

    def get_default_impl(
        self,
        base: Type[T],
        sort_key: Optional[Callable[[Type[T]], Any]] = None,
        reverse: bool = False,
        permit_no_impl: bool = False,
    ) -> Type:
        impls = self.get_impls(base, sort_key, reverse, permit_no_impl)
        if impls:
            return impls[0]

    # pylint: disable=R0913
    def get_instances(
        self,
        base: Type[T],
        sort_key: Optional[Callable[[Type[T]], Any]] = None,
        reverse: bool = False,
        kwargs: Optional[Dict] = None,
        permit_no_impl: bool = False,
    ) -> List[T]:
        if kwargs is None:
            kwargs = {}
        impls = self.get_impls(base, sort_key, reverse, permit_no_impl)
        result = [impl(**kwargs) for impl in impls]
        return result

    # pylint: disable=R0913
    def get_new_default_instance(
        self,
        base: Type[T],
        sort_key: Optional[Callable[[Type[T]], Any]] = None,
        reverse: bool = False,
        kwargs: Optional[Dict] = None,
        permit_no_impl: bool = False,
    ) -> T:
        if kwargs is None:
            kwargs = {}
        impls = self.get_impls(base, sort_key, reverse, permit_no_impl)
        if impls:
            result = impls[0](**kwargs)
            return result


def create_injecty_context(config_module_prefix: str = _CONFIG_MODULE_PREFIX):
    context = InjectyContext()
    module_info = (
        m for m in pkgutil.iter_modules() if m.name.startswith(config_module_prefix)
    )
    modules = [importlib.import_module(m.name) for m in module_info]
    modules.sort(key=lambda m: m.priority, reverse=False)
    for m in modules:
        getattr(m, "configure")(context)
    return context


def _priority_sort_key(n):
    return n.priority



import math
from dataclasses import dataclass

from example.models.shape_abc import ShapeABC


@dataclass
class Circle(ShapeABC):
    radius: float

    def get_area(self):
        return math.pi * self.radius**2



from typing import Type, Optional, Callable, Any, List, Dict

from injecty.injecty_context import InjectyContext, create_injecty_context, T

_DEFAULT_CONTEXT = None


def get_default_injecty_context():
    # pylint: disable=W0603
    global _DEFAULT_CONTEXT
    if _DEFAULT_CONTEXT is None:
        _DEFAULT_CONTEXT = create_injecty_context()
    return _DEFAULT_CONTEXT


def get_impls(
    base: Type[T],
    sort_key: Optional[Callable[[Type[T]], Any]] = None,
    reverse: bool = False,
    permit_no_impl: bool = False,
) -> List[Type]:
    context = get_default_injecty_context()
    impls = context.get_impls(base, sort_key, reverse, permit_no_impl)
    return impls


def get_instances(
    base: Type[T],
    sort_key: Optional[Callable[[Type[T]], Any]] = None,
    reverse: bool = False,
    kwargs: Optional[Dict] = None,
    permit_no_impl: bool = False,
) -> List[T]:
    context = get_default_injecty_context()
    instances = context.get_instances(base, sort_key, reverse, kwargs, permit_no_impl)
    return instances


def get_default_impl(
    base: Type[T],
    sort_key: Optional[Callable[[Type[T]], Any]] = None,
    reverse: bool = False,
    permit_no_impl: bool = False,
) -> Type:
    context = get_default_injecty_context()
    impl_ = context.get_default_impl(base, sort_key, reverse, permit_no_impl)
    return impl_


def get_new_default_instance(
    base: Type[T],
    sort_key: Optional[Callable[[Type[T]], Any]] = None,
    reverse: bool = False,
    kwargs: Optional[Dict] = None,
    permit_no_impl: bool = False,
) -> T:
    context = get_default_injecty_context()
    instance = context.get_new_default_instance(
        base, sort_key, reverse, kwargs, permit_no_impl
    )
    return instance



from injecty import InjectyContext
from tests.test_injecty_context import Foo, Bar, Zap, Bang

priority: int = 100


def configure(context: InjectyContext):
    context.register_impls(Foo, [Bar, Zap, Bang])



from unittest import TestCase

from injecty import InjectyContext, create_injecty_context
from tests.models import Foo, Bar, Bang, Zap


class TestInjectyContext(TestCase):
    def test_register_impl(self):
        context = InjectyContext()
        self.assertTrue(context.register_impl(Foo, Bar))
        impls = context.get_impls(Foo)
        self.assertEqual(impls, [Bar])

    def test_register_impl_unrelated_type(self):
        context = InjectyContext()
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            self.assertTrue(context.register_impl(int, str))
        # noinspection PyTypeChecker
        context.register_impl(int, float, False)
        impls = context.get_impls(int)
        self.assertEqual(impls, [float])

    def test_register_impl_already_registered(self):
        context = InjectyContext()
        self.assertTrue(context.register_impl(Foo, Bar))
        self.assertFalse(context.register_impl(Foo, Bar))
        impls = context.get_impls(Foo)
        self.assertEqual(impls, [Bar])

    def test_register_impls(self):
        context = InjectyContext()
        self.assertTrue(context.register_impls(Foo, [Bar, Zap, Bang]))
        impls = context.get_impls(Foo)
        self.assertEqual(impls, [Bang, Bar, Zap])

    def test_register_impls_already_registered(self):
        context = InjectyContext()
        context.register_impl(Foo, Zap)
        self.assertFalse(context.register_impls(Foo, [Bar, Zap, Bang]))
        impls = context.get_impls(Foo)
        self.assertEqual(impls, [Bang, Bar, Zap])

    def test_deregister_impl(self):
        context = InjectyContext()
        context.register_impls(Foo, [Bar, Zap, Bang])
        self.assertTrue(context.deregister_impl(Foo, Bar))
        impls = context.get_impls(Foo)
        self.assertEqual(impls, [Bang, Zap])

    def test_deregister_impl_not_registered(self):
        context = InjectyContext()
        context.register_impls(Foo, [Zap, Bang])
        self.assertFalse(context.deregister_impl(Foo, Bar))
        impls = context.get_impls(Foo)
        self.assertEqual(impls, [Bang, Zap])

    def test_load_from_module(self):
        context = create_injecty_context("injecty_config_test")
        impls = context.get_impls(Foo)
        self.assertEqual(impls, [Bang, Bar, Zap])

    def test_load_from_package(self):
        context = create_injecty_context("test_config_pac")
        impls = context.get_impls(Foo)
        self.assertEqual(impls, [Bang, Bar, Zap])

    def test_load_from_module_with_no_configure(self):
        with self.assertRaises(AttributeError):
            create_injecty_context("test_config_no_con")

    def test_load_from_module_with_no_priority(self):
        with self.assertRaises(AttributeError):
            create_injecty_context("test_config_no_pri")

    def test_get_impls(self):
        context = create_injecty_context("injecty_config_test")
        impls = context.get_impls(Foo, sort_key=lambda c: c.priority, reverse=True)
        self.assertEqual(impls, [Bang, Bar, Zap])
        impls = context.get_impls(Foo, sort_key=lambda c: c.priority)
        self.assertEqual(impls, [Zap, Bar, Bang])

    def test_get_impls_no_impl(self):
        context = InjectyContext()
        with self.assertRaises(ValueError):
            context.get_impls(int)

    def test_get_impls_no_impl_permitted(self):
        context = InjectyContext()
        self.assertEqual([], context.get_impls(int, permit_no_impl=True))

    def test_get_instances(self):
        context = create_injecty_context("injecty_config_test")
        instances = [i.__class__ for i in context.get_instances(Foo)]
        expected = [Bang, Bar, Zap]
        self.assertEqual(instances, expected)



from typing import Dict

from injecty import get_impls

from example.models.shape_abc import ShapeABC


def parse_shape(shape_dict: Dict) -> ShapeABC:
    shape_type = shape_dict.pop("type")
    impls = get_impls(ShapeABC)
    for impl in impls:
        if impl.__name__ == shape_type:
            return impl(**shape_dict)
    raise ValueError(f"no_implementation_for:{shape_type}")



import argparse
import json

from example.shape_parser import parse_shape

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Shape Areas")
    parser.add_argument("--input", required=True)
    args, _ = parser.parse_known_args()
    with open(args.input, "r") as input_file:
        shape_dicts = json.load(input_file)
        for shape_dict in shape_dicts:
            shape = parse_shape(shape_dict)
            print(shape.get_area())



from dataclasses import dataclass

from example.models.shape_abc import ShapeABC


@dataclass
class Rectangle(ShapeABC):
    length: float
    height: float

    def get_area(self):
        return self.length * self.height
